/*
 * manager-interfaces.c
 * This file is part of Network-inador
 *
 * Copyright (C) 2025 - Félix Arreola Rodríguez
 *
 * Network-inador is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Network-inador is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Network-inador; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 */

#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include <linux/if.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>

#include "manager-private.h"
#include "network-inador-public.h"
#include "network-inador-manager.h"

/* Empaquetador de datos */
int _manager_pack_interface_info (unsigned char *buffer, Interface *iface) {
	const unsigned char *iface_name;
	uint32_t u32; uint16_t u16; uint8_t u8;
	int string_len;
	
	iface_name = network_inador_iface_get_name (iface);
	string_len = strlen (iface_name);
	
	u32 = network_inador_iface_get_index (iface);
	memcpy (&buffer[0], &u32, 4);
	
	u32 = network_inador_iface_get_link_type (iface);
	memcpy (&buffer[4], &u32, 4);
	
	u32 = network_inador_iface_get_master_index (iface);
	memcpy (&buffer[8], &u32, 4);
	
	u32 = network_inador_iface_get_mtu (iface);
	memcpy (&buffer[12], &u32, 4);
	
	u16 = network_inador_iface_get_flags (iface);
	memcpy (&buffer[16], &u16, 2);
	
	buffer[18] = network_inador_iface_get_is_wireless (iface);
	buffer[19] = string_len;
	memcpy (&buffer[20], iface_name, string_len);
	
	return 20 + string_len;
}

/* Los comandos que verdaderamente envian */
void _manager_send_interface (ManagerClientInfo *manager_client, Interface *iface, uint8_t is_event) {
	unsigned char buffer[32 + IFNAMSIZ + IFNAMSIZ];
	int size;
	
	if (is_event) {
		buffer[0] = NET_INADOR_TYPE_EVENT;
		buffer[1] = NET_INADOR_EVENT_IFACE_ADDED;
	} else {
		buffer[0] = NET_INADOR_TYPE_RESPONSE;
		buffer[1] = NET_INADOR_RESPONSE_IFACE;
	}
	
	size = 2 + _manager_pack_interface_info (&buffer[2], iface);
	
	send (manager_client->fd, buffer, size, 0);
}

void _manager_send_interface_deleted (ManagerClientInfo *manager_client, uint32_t index) {
	unsigned char buffer[32];
	
	buffer[0] = NET_INADOR_TYPE_EVENT;
	buffer[1] = NET_INADOR_EVENT_IFACE_REMOVED;
	
	memcpy (&buffer[2], &index, 4);
	
	send (manager_client->fd, buffer, 6, 0);
}

void _manager_interface_handle_list_interfaces (ManagerClientInfo *manager_client, unsigned char *buffer, int buffer_len) {
	uint32_t *list_index;
	int g;
	Interface *iface;
	
	list_index = network_inador_list_ifaces (manager_client->manager->handle);
	
	/* TODO Analizar si podemos hacer todo en un solo envio incluyendo el comando de fin de listado */
	if (list_index == NULL) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_UNKNOWN, NET_INADOR_COMMAND_LIST_IFACES);
		
		return;
	}
	
	g = 0;
	while (list_index[g] != 0) {
		iface = network_inador_get_iface (manager_client->manager->handle, list_index[g]);
		if (iface == NULL) continue;
		
		_manager_send_interface (manager_client, iface, FALSE);
		
		g++;
	}
	
	_manager_send_end_command (manager_client, NET_INADOR_COMMAND_LIST_IFACES);
	free (list_index);
}

void _manager_interface_handle_get_interface (ManagerClientInfo *manager_client, unsigned char *buffer, int buffer_len) {
	Interface *iface;
	uint32_t u32;
	
	if (buffer_len < 6) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INCOMPLETE_REQUEST, NET_INADOR_COMMAND_GET_IFACE);
		return;
	}
	
	memcpy (&u32, &buffer[2], 4);
	iface = network_inador_get_iface (manager_client->manager->handle, u32);
	if (iface == NULL) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INVALID_IFACE_INDEX, NET_INADOR_COMMAND_GET_IFACE);
		return;
	}
	
	_manager_send_interface (manager_client, iface, FALSE);
}

void _manager_interface_handle_up_down_command (ManagerClientInfo *manager_client, uint8_t is_up, unsigned char *buffer, int buffer_len) {
	Interface *iface;
	uint32_t u32;
	int ret;
	
	if (buffer_len < 6) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INCOMPLETE_REQUEST, (is_up ? NET_INADOR_COMMAND_IFACE_UP : NET_INADOR_COMMAND_IFACE_DOWN));
		return;
	}
	
	memcpy (&u32, &buffer[2], 4);
	iface = network_inador_get_iface (manager_client->manager->handle, u32);
	if (iface == NULL) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INVALID_IFACE_INDEX, (is_up ? NET_INADOR_COMMAND_IFACE_UP : NET_INADOR_COMMAND_IFACE_DOWN));
		return;
	}
	
	if (is_up) {
		ret = interfaces_change_set_up (manager_client->manager->handle, u32);
	} else {
		ret = interfaces_change_set_down (manager_client->manager->handle, u32);
	}
	
	if (ret == 0) {
		/* OK */
		_manager_send_executed (manager_client);
	} else {
		_manager_send_error (manager_client, NET_INADOR_ERROR_NOT_EXECUTED, (is_up ? NET_INADOR_COMMAND_IFACE_UP : NET_INADOR_COMMAND_IFACE_DOWN));
	}
}

void _manager_interface_handle_change_name (ManagerClientInfo *manager_client, unsigned char *buffer, int buffer_len) {
	Interface *iface;
	uint32_t u32;
	int ret;
	int name_len;
	unsigned char name[IFNAMSIZ];
	
	if (buffer_len < 7) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INCOMPLETE_REQUEST, NET_INADOR_COMMAND_IFACE_CHANGE_NAME);
		return;
	}
	
	memcpy (&u32, &buffer[2], 4);
	iface = network_inador_get_iface (manager_client->manager->handle, u32);
	if (iface == NULL) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INVALID_IFACE_INDEX, NET_INADOR_COMMAND_IFACE_CHANGE_NAME);
		return;
	}
	
	name_len = buffer[6];
	if (name_len == 0 || name_len >= IFNAMSIZ) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_BAD_STRING, NET_INADOR_COMMAND_IFACE_CHANGE_NAME);
		return;
	}
	
	if (name_len + 7 < buffer_len) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INCOMPLETE_REQUEST, NET_INADOR_COMMAND_IFACE_CHANGE_NAME);
		return;
	}
	
	memcpy (name, &buffer[7], name_len);
	if (name[0] == 0) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_BAD_STRING, NET_INADOR_COMMAND_IFACE_CHANGE_NAME);
		return;
	}
	name[name_len] = 0;
	
	ret = interfaces_change_name (manager_client->manager->handle, u32, name);
	
	if (ret == 0) {
		/* OK */
		_manager_send_executed (manager_client);
	} else {
		_manager_send_error (manager_client, NET_INADOR_ERROR_NOT_EXECUTED, NET_INADOR_COMMAND_IFACE_CHANGE_NAME);
	}
}

void _manager_interface_handle_change_mtu (ManagerClientInfo *manager_client, unsigned char *buffer, int buffer_len) {
	Interface *iface;
	uint32_t u32, new_mtu;
	int ret;
	int name_len;
	unsigned char name[IFNAMSIZ];
	
	if (buffer_len < 10) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INCOMPLETE_REQUEST, NET_INADOR_COMMAND_IFACE_CHANGE_MTU);
		return;
	}
	
	memcpy (&u32, &buffer[2], 4);
	iface = network_inador_get_iface (manager_client->manager->handle, u32);
	if (iface == NULL) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INVALID_IFACE_INDEX, NET_INADOR_COMMAND_IFACE_CHANGE_MTU);
		return;
	}
	
	memcpy (&new_mtu, &buffer[6], 4);
	//new_mtu = ntohl (new_mtu);
	
	/* TODO: Revisar el máximo de MTU que se puede poner */
	ret = interfaces_change_mtu (manager_client->manager->handle, u32, new_mtu);
	
	if (ret == 0) {
		/* OK */
		_manager_send_executed (manager_client);
	} else {
		_manager_send_error (manager_client, NET_INADOR_ERROR_NOT_EXECUTED, NET_INADOR_COMMAND_IFACE_CHANGE_MTU);
	}
}

/* Los eventos que vienen desde la librería */
void _manager_interface_added_event_cb (NetworkInadorHandle *handle, Interface *iface, void *data) {
	Manager *manager = (Manager *) data;
	GList *g;
	ManagerClientInfo *manager_client;
	
	printf ("___ MANAGER ___ Informando interfaz agregada: %s (%i)\n", network_inador_iface_get_name (iface), network_inador_iface_get_index (iface));
	
	for (g = manager->connected_client_list; g != NULL; g = g->next) {
		manager_client = (ManagerClientInfo *) g->data;
		
		if (manager_client->wanted_events & NET_INADOR_EVENT_MASK_INTERFACES) {
			printf ("___ MANAGER ___ Informando a la conexión (%i)\n", manager_client->fd);
			_manager_send_interface (manager_client, iface, TRUE);
		}
	}
}

void _manager_interface_updated_event_cb (NetworkInadorHandle *handle, Interface *iface, void *data) {
	Manager *manager = (Manager *) data;
	GList *g;
	ManagerClientInfo *manager_client;
	
	printf ("___ MANAGER ___ Informando interfaz actualizada: %s (%i)\n", network_inador_iface_get_name (iface), network_inador_iface_get_index (iface));
	
	for (g = manager->connected_client_list; g != NULL; g = g->next) {
		manager_client = (ManagerClientInfo *) g->data;
		
		if (manager_client->wanted_events & NET_INADOR_EVENT_MASK_INTERFACES) {
			printf ("___ MANAGER ___ Informando a la conexión (%i)\n", manager_client->fd);
			_manager_send_interface (manager_client, iface, TRUE);
		}
	}
}

void _manager_interface_deleted_event_cb (NetworkInadorHandle *handle, uint32_t index, void *data) {
	Manager *manager = (Manager *) data;
	GList *g;
	ManagerClientInfo *manager_client;
	
	printf ("___ MANAGER ___ Informando interfaz eliminada: %i\n", index);
	
	for (g = manager->connected_client_list; g != NULL; g = g->next) {
		manager_client = (ManagerClientInfo *) g->data;
		
		if (manager_client->wanted_events & NET_INADOR_EVENT_MASK_INTERFACES) {
			printf ("___ MANAGER ___ Informando a la conexión (%i)\n", manager_client->fd);
			_manager_send_interface_deleted (manager_client, index);
		}
	}
}

