/*
 * manager-ip.c
 * This file is part of Network-inador
 *
 * Copyright (C) 2025 - Félix Arreola Rodríguez
 *
 * Network-inador is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Network-inador is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Network-inador; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 */

#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include <linux/if.h>
#include <linux/if_addr.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>

#include "manager-private.h"
#include "network-inador-public.h"
#include "network-inador-manager.h"

int _manager_pack_ip_info (unsigned char *buffer, IPAddr *ip_addr) {
	int family_size = 0;
	int label_len;
	int pos;
	int family;
	uint32_t index, flags;
	Interface *iface;
	int has_brd, has_local;
	struct ifa_cacheinfo cacheinfo;
	const char *label;
	
	union {
		struct in_addr v4;
		struct in6_addr v6;
	} addr;
	
	/* Familia y máscara */
	family = network_inador_ipaddr_get_family (ip_addr);
	if (family == AF_INET) {
		family_size = sizeof (struct in_addr);
	} else if (family == AF_INET6) {
		family_size = sizeof (struct in6_addr);
	}
	
	iface = network_inador_ipaddr_get_iface (ip_addr);
	index = network_inador_iface_get_index (iface);
	memcpy (&buffer[0], &index, 4);
	buffer[4] = family;
	buffer[5] = network_inador_ipaddr_get_prefix (ip_addr);
	
	has_local = network_inador_ipaddr_has_local (ip_addr);
	has_brd = network_inador_ipaddr_has_brd (ip_addr);
	
	/* Los campos de bits */
	buffer[6] = 0;
	if (has_local) {
		buffer[6] |= 0x01;
	}
	if (has_brd) {
		buffer[6] |= 0x02;
	}
	
	/* Byte scope y las flags */
	buffer[7] = network_inador_ipaddr_get_scope (ip_addr);
	flags = network_inador_ipaddr_get_flags (ip_addr);
	memcpy (&buffer[8], &flags, 4);
	
	network_inador_ipaddr_get_cacheinfo (ip_addr, &cacheinfo);
	/* La información de marcas de tiempo de las IPs */
	memcpy (&buffer[12], &cacheinfo, 16);
	
	/* La dirección principal */
	network_inador_ipaddr_get_addr (ip_addr, &addr, NULL);
	memcpy (&buffer[28], &addr, family_size);
	pos = 28 + family_size;
	if (has_local) {
		network_inador_ipaddr_get_local_addr (ip_addr, &addr, NULL);
		memcpy (&buffer[pos], &addr, family_size);
		pos += family_size;
	}
	
	if (has_brd) {
		network_inador_ipaddr_get_brd_addr (ip_addr, &addr, NULL);
		memcpy (&buffer[pos], &addr, family_size);
		pos += family_size;
	}
	
	label = network_inador_ipaddr_get_label (ip_addr);
	if (label[0] != 0) {
		label_len = strlen (label);
		buffer[pos] = label_len;
		pos++;
		
		memcpy (&buffer[pos], &label, label_len);
		pos += label_len;
	} else {
		buffer[pos] = 0;
		pos++;
	}
	
	return pos;
}

/* Los comandos que verdaderamente envian */
void _manager_send_ipaddr (ManagerClientInfo *manager_client, IPAddr *ip_addr, uint8_t is_event) {
	unsigned char buffer[80];
	int size;
	
	if (is_event) {
		buffer[0] = NET_INADOR_TYPE_EVENT;
		buffer[1] = NET_INADOR_EVENT_IPADDR_ADDED;
	} else {
		buffer[0] = NET_INADOR_TYPE_RESPONSE;
		buffer[1] = NET_INADOR_RESPONSE_IPADDR;
	}
	
	size = 2 + _manager_pack_ip_info (&buffer[2], ip_addr);
	
	send (manager_client->fd, buffer, size, 0);
}

void _manager_send_ipaddr_deleted (ManagerClientInfo *manager_client, IPAddr *ip_addr) {
	unsigned char buffer[80];
	int family_size = 0;
	int pos;
	int family;
	uint32_t index;
	Interface *iface;
	int has_brd, has_local;
	
	union {
		struct in_addr v4;
		struct in6_addr v6;
	} addr;
	
	buffer[0] = NET_INADOR_TYPE_EVENT;
	buffer[1] = NET_INADOR_EVENT_IPADDR_REMOVED;
	
	/* Familia y máscara */
	family = network_inador_ipaddr_get_family (ip_addr);
	if (family == AF_INET) {
		family_size = sizeof (struct in_addr);
	} else if (family == AF_INET6) {
		family_size = sizeof (struct in6_addr);
	}
	iface = network_inador_ipaddr_get_iface (ip_addr);
	index = network_inador_iface_get_index (iface);
	memcpy (&buffer[2], &index, 4);
	buffer[6] = family;
	buffer[7] = network_inador_ipaddr_get_prefix (ip_addr);
	
	has_local = network_inador_ipaddr_has_local (ip_addr);
	
	buffer[8] = 0;
	if (has_local) {
		buffer[8] |= 0x01;
	}
	
	buffer[9] = 0;
	
	network_inador_ipaddr_get_addr (ip_addr, &addr, NULL);
	memcpy (&buffer[10], &addr, family_size);
	pos = 10 + family_size;
	if (has_local) {
		network_inador_ipaddr_get_local_addr (ip_addr, &addr, NULL);
		memcpy (&buffer[pos], &addr, family_size);
		pos += family_size;
	}
	
	send (manager_client->fd, buffer, pos, 0);
}

void _manager_ip_handle_list_ips (ManagerClientInfo *manager_client, unsigned char *buffer, int buffer_len) {
	Interface *iface;
	int family, ip_family;
	uint32_t u32;
	int c, total;
	IPAddr *ip_addr;
	
	if (buffer_len < 7) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INCOMPLETE_REQUEST, NET_INADOR_COMMAND_LIST_IP);
		return;
	}
	
	family = buffer[6];
	
	if (family != AF_UNSPEC && family != AF_INET && family != AF_INET6) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INVALID_FAMILY, NET_INADOR_COMMAND_LIST_IP);
		
		return;
	}
	
	memcpy (&u32, &buffer[2], 4);
	iface = network_inador_get_iface (manager_client->manager->handle, u32);
	if (iface == NULL) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INVALID_IFACE_INDEX, NET_INADOR_COMMAND_LIST_IP);
		return;
	}
	
	total = network_inador_iface_get_num_ips (iface);
	for (c = 0; c < total; c++) {
		ip_addr = network_inador_iface_get_ipaddr_by_index (iface, c);
		
		ip_family = network_inador_ipaddr_get_family (ip_addr);
		
		if (family != AF_UNSPEC && family != ip_family) continue;
		
		_manager_send_ipaddr (manager_client, ip_addr, FALSE);
	}
	
	_manager_send_end_command (manager_client, NET_INADOR_COMMAND_LIST_IP);
}

void _manager_ip_handle_clear_ips (ManagerClientInfo *manager_client, unsigned char *buffer, int buffer_len) {
	int family, ip_family;
	Interface *iface;
	IPAddr *ip_addr;
	uint32_t u32;
	int ret;
	int c, total;
	
	if (buffer_len < 7) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INCOMPLETE_REQUEST, NET_INADOR_COMMAND_CLEAR_IP);
		return;
	}
	
	family = buffer[6];
	
	if (family != AF_UNSPEC && family != AF_INET && family != AF_INET6) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INVALID_FAMILY, NET_INADOR_COMMAND_CLEAR_IP);
		
		return;
	}
	
	memcpy (&u32, &buffer[2], 4);
	iface = network_inador_get_iface (manager_client->manager->handle, u32);
	if (iface == NULL) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INVALID_IFACE_INDEX, NET_INADOR_COMMAND_CLEAR_IP);
		return;
	}
	
	ret = 0;
	total = network_inador_iface_get_num_ips (iface);
	for (c = 0; c < total; c++) {
		ip_addr = network_inador_iface_get_ipaddr_by_index (iface, c);
		
		ip_family = network_inador_ipaddr_get_family (ip_addr);
		
		if (family != AF_UNSPEC && family != ip_family) continue;
		
		// AQUÍ GATUNO
		//ret |= ip_address_del_ip (manager_client->manager->handle, iface->index, ip_addr);
	}
	
	if (ret == 0) {
		/* OK */
		_manager_send_executed (manager_client);
	} else {
		_manager_send_error (manager_client, NET_INADOR_ERROR_NOT_EXECUTED, NET_INADOR_COMMAND_CLEAR_IP);
	}
}

/* Los eventos que vienen desde la librería */
void _manager_ip_added_event_cb (NetworkInadorHandle *handle, IPAddr *ip_addr, void *data) {
	Manager *manager = (Manager *) data;
	GList *g;
	ManagerClientInfo *manager_client;
	Interface *iface;
	
	iface = network_inador_ipaddr_get_iface (ip_addr);
	printf ("___ MANAGER ___ Informando ip agregada: %s (%i)\n", network_inador_iface_get_name (iface), network_inador_iface_get_index (iface));
	
	for (g = manager->connected_client_list; g != NULL; g = g->next) {
		manager_client = (ManagerClientInfo *) g->data;
		
		if (manager_client->wanted_events & NET_INADOR_EVENT_MASK_INTERFACES) {
			printf ("___ MANAGER ___ Informando a la conexión (%i)\n", manager_client->fd);
			_manager_send_ipaddr (manager_client, ip_addr, TRUE);
		}
	}
}

void _manager_ip_deleted_event_cb (NetworkInadorHandle *handle, IPAddr *ip_addr, void *data) {
	Manager *manager = (Manager *) data;
	GList *g;
	ManagerClientInfo *manager_client;
	Interface *iface;
	
	iface = network_inador_ipaddr_get_iface (ip_addr);
	printf ("___ MANAGER ___ Informando ip eliminada: %s (%i)\n", network_inador_iface_get_name (iface), network_inador_iface_get_index (iface));
	
	for (g = manager->connected_client_list; g != NULL; g = g->next) {
		manager_client = (ManagerClientInfo *) g->data;
		
		if (manager_client->wanted_events & NET_INADOR_EVENT_MASK_IP) {
			printf ("___ MANAGER ___ Informando a la conexión (%i)\n", manager_client->fd);
			_manager_send_ipaddr_deleted (manager_client, ip_addr);
		}
	}
}

