/*
 * manager.c
 * This file is part of Network-inador
 *
 * Copyright (C) 2024 - Félix Arreola Rodríguez
 *
 * Network-inador is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Network-inador is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Network-inador; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 */

#include <glib.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>

#include <unistd.h>
#include <fcntl.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>

#include <sys/un.h>

#include <linux/if_addr.h>
#include <sys/stat.h>

#include "manager-private.h"
#include "network-inador-public.h"
#include "network-inador-events.h"
#include "network-inador-manager.h"

#include "manager-interfaces.h"
#include "manager-ip.h"

#define COMMAND_SOCKET_PATH "/tmp/network-inador.socket"

/* Funciones comúnes */
void _manager_send_error (ManagerClientInfo *manager_client, int error, int orig_cmd) {
	unsigned char buffer[8];
	
	buffer[0] = NET_INADOR_TYPE_RESPONSE_ERROR;
	buffer[1] = orig_cmd;
	buffer[2] = error;
	
	send (manager_client->fd, buffer, 3, 0);
}

void _manager_send_executed (ManagerClientInfo *manager_client) {
	unsigned char buffer[8];
	
	buffer[0] = NET_INADOR_TYPE_RESPONSE;
	buffer[1] = NET_INADOR_RESPONSE_EXECUTED;
	
	send (manager_client->fd, buffer, 2, 0);
}

void _manager_send_end_command (ManagerClientInfo *manager_client, int orig_cmd) {
	unsigned char buffer[8];
	
	buffer[0] = NET_INADOR_TYPE_RESPONSE_LISTING_END;
	buffer[1] = orig_cmd;
	
	send (manager_client->fd, buffer, 2, 0);
}

static void _manager_handle_set_event_mask (ManagerClientInfo *manager_client, unsigned char *buffer, int buffer_len) {
	uint32_t events;
	
	if (buffer_len < 6) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INCOMPLETE_REQUEST, NET_INADOR_COMMAND_SET_EVENT_MASK);
		return;
	}
	
	memcpy (&events, &buffer[2], 4);
	printf ("___ MANAGER __ Set events de un cliente (%i), eventos: %u\n", manager_client->fd, events);
	
	manager_client->wanted_events = events;
}

/* Los eventos que vienen desde la librería */
void _manager_dhcp_event_cb (NetworkInadorHandle *handle, Interface *iface, InterfaceDHCPClientInfo *dhcp_info, void *data) {
	Manager *manager = (Manager *) data;
}


static gboolean _manager_on_client_data_recv (GIOChannel *source, GIOCondition condition, gpointer data) {
	ManagerClientInfo *manager_client = (ManagerClientInfo *) data;
	Manager *manager = manager_client->manager;
	int type, command;
	Interface *iface;
	int g;
	
	unsigned char buffer[8192];
	int bytes;
	
	bytes = recv (manager_client->fd, buffer, sizeof (buffer), 0);
	
	if (bytes <= 0) {
		printf ("___ MANAGER ___ Conexión cerrada manager (%i)\n", manager_client->fd);
		/* Error de lectura o cierre */
		close (manager_client->fd);
		
		manager->connected_client_list = g_list_remove (manager->connected_client_list, manager_client);
		free (manager_client);
		
		return FALSE;
	}
	
	printf ("___ MANAGER ___ Data arrival (%i):\n", manager_client->fd);
	for (g = 0; g < bytes; g++) {
		printf ("%02hhx ", buffer[g]);
	}
	printf ("\n");
	
	/* Comando incompleto */
	if (bytes < 2) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INCOMPLETE_REQUEST, 0);
		
		return TRUE;
	}
	
	/* Procesar los datos del cliente */
	type = buffer[0];
	if (type != NET_INADOR_TYPE_COMMAND) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_WRONG_COMMAND, 0);
		
		return TRUE;
	}
	
	command = buffer[1];
	
	switch (command) {
		case NET_INADOR_COMMAND_LIST_IFACES:
			_manager_interface_handle_list_interfaces (manager_client, buffer, bytes);
			break;
		case NET_INADOR_COMMAND_GET_IFACE:
			_manager_interface_handle_get_interface (manager_client, buffer, bytes);
			break;
		case NET_INADOR_COMMAND_IFACE_UP:
		case NET_INADOR_COMMAND_IFACE_DOWN:
			_manager_interface_handle_up_down_command (manager_client, (command == NET_INADOR_COMMAND_IFACE_UP), buffer, bytes);
			break;
		case NET_INADOR_COMMAND_IFACE_CHANGE_NAME:
			_manager_interface_handle_change_name (manager_client, buffer, bytes);
			break;
		case NET_INADOR_COMMAND_IFACE_CHANGE_MTU:
			_manager_interface_handle_change_mtu (manager_client, buffer, bytes);
			break;
		case NET_INADOR_COMMAND_LIST_IP:
			_manager_ip_handle_list_ips (manager_client, buffer, bytes);
			break;
		/*case NET_INADOR_COMMAND_CLEAR_IP:
			_manager_execute_clear_ips (manager_client, buffer, bytes);
			break;
		case NET_INADOR_COMMAND_ADD_IP:
			_manager_execute_add_ip (manager_client, buffer, bytes);
			break;
		case NET_INADOR_COMMAND_REMOVE_IP:
			_manager_execute_delete_ip (manager_client, buffer, bytes);
			break;
		*/
		case NET_INADOR_COMMAND_SET_EVENT_MASK:
			_manager_handle_set_event_mask (manager_client, buffer, bytes);
			break;
		/*case NET_INADOR_COMMAND_CREATE_BRIDGE:
			_manager_execute_create_bridge (manager_client, buffer, bytes);
			break;
		case NET_INADOR_COMMAND_CLEAR_MASTER:
			_manager_execute_set_or_clear_master (manager_client, buffer, bytes, TRUE);
			break;
		case NET_INADOR_COMMAND_SET_MASTER:
			_manager_execute_set_or_clear_master (manager_client, buffer, bytes, FALSE);
			break;
		case NET_INADOR_COMMAND_DHCP_CLIENT_FEED:
			_manager_execute_dhcp_client_feed (manager_client, buffer, bytes);
			break;
		case NET_INADOR_COMMAND_RUN_DHCP:
			_manager_execute_dhcp_run (manager_client, buffer, bytes);
			break;
		case NET_INADOR_COMMAND_STOP_DHCP:
			_manager_execute_dhcp_stop (manager_client, buffer, bytes);
			break;
		case NET_INADOR_COMMAND_GET_DHCP_STATUS:
			_manager_execute_dhcp_get_status (manager_client, buffer, bytes);
			break;
		case NET_INADOR_COMMAND_LIST_ROUTES:
			_manager_send_list_routes (manager_client, buffer, bytes);
			break;
		case NET_INADOR_COMMAND_ADD_ROUTE:
			_manager_execute_add_route (manager_client, buffer, bytes);
			break;
		case NET_INADOR_COMMAND_REMOVE_ROUTE:
			_manager_execute_del_route (manager_client, buffer, bytes);
			break;
		case NET_INADOR_COMMAND_LIST_ROUTE_TABLES:
			_manager_send_list_route_tables (manager_client, buffer, bytes);
			break;
		case NET_INADOR_COMMAND_RESOLVCONF_FEED:
			_manager_execute_resolvconf_feed (manager_client, buffer, bytes);
			break;
		case NET_INADOR_COMMAND_RESOLVCONF_REMOVE:
			_manager_execute_resolvconf_remove (manager_client, buffer, bytes);
			break;*/
		default:
			_manager_send_error (manager_client, NET_INADOR_ERROR_WRONG_COMMAND, command);
	}
	
	
	return TRUE;
}

static gboolean _manager_on_client_connect (GIOChannel *source, GIOCondition condition, gpointer data) {
	Manager *manager = (Manager *) data;
	ManagerClientInfo *manager_client;
	int fd;
	GIOChannel *channel;
	
	fd = accept (manager->socket, NULL, NULL);
	printf ("___ MANAGER ___ Nueva conexión (%i)\n", fd);
	
	fcntl (fd, F_SETFD, 1);
	manager_client = malloc (sizeof (ManagerClientInfo));
	
	if (manager_client == NULL) {
		close (fd);
		
		return TRUE;
	}
	
	memset (manager_client, 0, sizeof (*manager_client));
	
	manager_client->fd = fd;
	
	/* Generar el vigilador */
	channel = g_io_channel_unix_new (fd);
	manager_client->source = g_io_add_watch (channel, G_IO_IN | G_IO_PRI | G_IO_ERR | G_IO_HUP, _manager_on_client_data_recv, manager_client);
	g_io_channel_unref (channel);
	manager_client->manager = manager;
	
	manager->connected_client_list = g_list_prepend (manager->connected_client_list, manager_client);
	
	return TRUE;
}

Manager * manager_new (NetworkInadorHandle *handle, int type) {
	Manager *manager;
	struct sockaddr_un socket_name;
	GIOChannel *channel;
	
	manager = malloc (sizeof (Manager));
	if (manager == NULL) {
		return NULL;
	}
	
	memset (manager, 0, sizeof (*manager));
	
	/* TODO: Aquí revisar el tipo de socket a abrir */
	manager->socket = socket (AF_UNIX, SOCK_SEQPACKET | SOCK_CLOEXEC, 0);
	
	if (manager->socket < 0) {
		perror ("Failed to create AF_UNIX socket");
		
		free (manager);
		
		return NULL;
	}
	
	fcntl (manager->socket, F_SETFD, 1);
	
	memset (&socket_name, 0, sizeof (struct sockaddr_un));
	
	socket_name.sun_family = AF_UNIX;
	strncpy (socket_name.sun_path, COMMAND_SOCKET_PATH, sizeof (socket_name.sun_path) - 1);
	
	unlink (COMMAND_SOCKET_PATH);
	
	if (bind (manager->socket, (struct sockaddr *) &socket_name, sizeof (struct sockaddr_un)) < 0) {
		perror ("bind");
		
		close (manager->socket);
		manager->socket = -1;
		free (manager);
		
		return NULL;
	}
	
	if (listen (manager->socket, 10) < 0) {
		perror ("listen");
		
		close (manager->socket);
		manager->socket = -1;
		free (manager);
		
		return NULL;
	}
	
	/* TODO: Aplicar permisos aquí */
	chmod (COMMAND_SOCKET_PATH, 0666);
	
	channel = g_io_channel_unix_new (manager->socket);
	
	manager->source = g_io_add_watch (channel, G_IO_IN | G_IO_PRI | G_IO_ERR | G_IO_HUP, _manager_on_client_connect, manager);
	g_io_channel_unref (channel);
	manager->handle = handle;
	
	/* Agregar vigilancias aquí */
	network_manager_add_watch_dhcp (handle, _manager_dhcp_event_cb, manager);
	network_manager_add_watch_interface_added (handle, _manager_interface_added_event_cb, manager);
	network_manager_add_watch_interface_updated (handle, _manager_interface_updated_event_cb, manager);
	network_manager_add_watch_interface_deleted (handle, _manager_interface_deleted_event_cb, manager);
	network_manager_add_watch_ip_added (handle, _manager_ip_added_event_cb, manager);
	network_manager_add_watch_ip_deleted (handle, _manager_ip_deleted_event_cb, manager);
	
	return manager;
}

